/*----------------------------------------------------------------------------
    calories - real-time, networked, cellular automaton simulation

    Copyright (c) 2008 Ge Wang
        All rights reserved.
        http://ccrma.stanford.edu/~ge/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    U.S.A.
-----------------------------------------------------------------------------*/

//-----------------------------------------------------------------------------
// name: ca.cpp
// desc: implementation of cellular automaton classes
//
// author: Ge Wang (ge@ccrma.stanford.edu)
//   date: 2008.1.15
//-----------------------------------------------------------------------------
#include "ca.h"




//-----------------------------------------------------------------------------
// name: update()
// desc: ...
//-----------------------------------------------------------------------------
void CA::update()
{
    long x, y, z;
    Cell *** temp, * curr, * next2;

    // copy
    copy( next, grid );
    // reset
    numActive = 0;

    // go
    for( x = 0; x < width; x++ )
    {
        // go
        for( y = 0; y < height; y++ )
        {
            // go
            for( z = 0; z < depth; z++ )
            {
                // apply rules
                curr = &(grid[x][y][z]); next2 = &(next[x][y][z]);
                this->update( curr, next2 );
                // transitions
                if( curr->value > 0 && next2->value == 0 ) next2->sizeto = 0;
                else if( curr->value == 0 && next2->value > 0 ) next2->sizeto = 1;
                // count
                if( curr->value ) numActive++;
            }
        }
    }

    // copy
    copy( grid, next );

    // swap
    temp = grid;
    grid = next;
    next = temp;
}




//-----------------------------------------------------------------------------
// name: updateLook()
// desc: ...
//-----------------------------------------------------------------------------
void CA::updateLook()
{
    long x, y, z;
    Cell * curr;

    // go
    for( x = 0; x < width; x++ )
    {
        // go
        for( y = 0; y < height; y++ )
        {
            // go
            for( z = 0; z < depth; z++ )
            {
                curr = &(grid[x][y][z]);
                curr->size = (curr->sizeto-curr->size)*slew + curr->size;
            }
        }
    }
}




//-----------------------------------------------------------------------------
// name: getCell()
// desc: ...
//-----------------------------------------------------------------------------
const Cell * CA::getCell( const Cell * from, long dx, long dy, long dz )
{
    // target
    long x = from->x + dx, y = from->y + dy, z = from->z + dz;

    // wrap
    x %= width; if( x < 0 ) x += width;
    y %= height; if( y < 0 ) y += height;
    z %= depth; if( z < 0 ) z += depth;

    return &(grid[x][y][z]);
}




//-----------------------------------------------------------------------------
// name: getCell()
// desc: ...
//-----------------------------------------------------------------------------
const Cell * CA::getCell( long x, long y, long z )
{
    // check
    if( x < 0 || x >= width ) return NULL;
    if( y < 0 || y >= height ) return NULL;
    if( z < 0 || z >= depth ) return NULL;

    return &(grid[x][y][z]);
}




//-----------------------------------------------------------------------------
// name: setCell()
// desc: ...
//-----------------------------------------------------------------------------
bool CA::setCell( long x, long y, long z, long value )
{
    // check
    if( x < 0 || x >= width ) return false;
    if( y < 0 || y >= height ) return false;
    if( z < 0 || z >= depth ) return false;

    // get cell
    Cell * cell = &grid[x][y][z];

    // set state
    cell->value = value;
    // set sizeto
    cell->sizeto = (value != 0);

    return true;
}




//-----------------------------------------------------------------------------
// name: getNumNeighbors()
// desc: ...
//-----------------------------------------------------------------------------
long CA::getNumNeighbors( const Cell * from )
{
    long count = 0;
    const Cell * cell;
    // count
    cell = getCell( from, LEFT, 0, 0 ); count += (cell->value != 0);
    cell = getCell( from, LEFT, BELOW, 0 ); count += (cell->value != 0);
    cell = getCell( from, 0, BELOW, 0 ); count += (cell->value != 0);
    cell = getCell( from, RIGHT, BELOW, 0 ); count += (cell->value != 0);
    cell = getCell( from, RIGHT, 0, 0 ); count += (cell->value != 0);
    cell = getCell( from, RIGHT, ABOVE, 0 ); count += (cell->value != 0);
    cell = getCell( from, 0, ABOVE, 0 ); count += (cell->value != 0);
    cell = getCell( from, LEFT, ABOVE, 0 ); count += (cell->value != 0);

    // for 3D only
    if( depth > 1 )
    {
        cell = getCell( from, LEFT, 0, TOWARD ); count += (cell->value != 0);
        cell = getCell( from, LEFT, BELOW, TOWARD ); count += (cell->value != 0);
        cell = getCell( from, 0, BELOW, TOWARD ); count += (cell->value != 0);
        cell = getCell( from, RIGHT, BELOW, TOWARD ); count += (cell->value != 0);
        cell = getCell( from, RIGHT, 0, TOWARD ); count += (cell->value != 0);
        cell = getCell( from, RIGHT, ABOVE, TOWARD ); count += (cell->value != 0);
        cell = getCell( from, 0, ABOVE, TOWARD ); count += (cell->value != 0);
        cell = getCell( from, LEFT, ABOVE, TOWARD ); count += (cell->value != 0);
        cell = getCell( from, 0, 0, TOWARD ); count += (cell->value != 0);

        cell = getCell( from, LEFT, 0, AWAY ); count += (cell->value != 0);
        cell = getCell( from, LEFT, BELOW, AWAY ); count += (cell->value != 0);
        cell = getCell( from, 0, BELOW, AWAY ); count += (cell->value != 0);
        cell = getCell( from, RIGHT, BELOW, AWAY ); count += (cell->value != 0);
        cell = getCell( from, RIGHT, 0, AWAY ); count += (cell->value != 0);
        cell = getCell( from, RIGHT, ABOVE, AWAY ); count += (cell->value != 0);
        cell = getCell( from, 0, ABOVE, AWAY ); count += (cell->value != 0);
        cell = getCell( from, LEFT, ABOVE, AWAY ); count += (cell->value != 0);
        cell = getCell( from, 0, 0, AWAY ); count += (cell->value != 0);
    }

    return count;
}




//-----------------------------------------------------------------------------
// name: randomize()
// desc: ...
//-----------------------------------------------------------------------------
void CA::randomize( double prob )
{
    long x, y, z;

    // go
    for( x = 0; x < width; x++ )
    {
        for( y = 0; y < height; y++ )
        {
            for( z = 0; z < depth; z++ )
            {
                grid[x][y][z].value = (rand()/(double)RAND_MAX < prob );
                grid[x][y][z].sizeto = grid[x][y][z].value;
            }
        }
    }
}
