/*----------------------------------------------------------------------------
    calories - real-time, networked, 3D cellular automaton simulation

    Copyright (c) 2008 Ge Wang
        All rights reserved.
        http://ccrma.stanford.edu/~ge/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    U.S.A.
-----------------------------------------------------------------------------*/

//-----------------------------------------------------------------------------
// name: calories.cpp
// desc: simulator
//
// authors: Ge Wang (ge@ccrma.stanford.edu)
// library:
//          (OscPack) Ross Becina
//
// usage: (type 'h' while running or see list of command line arguments)
//-----------------------------------------------------------------------------
#include <math.h>
#include <stdio.h>

// global def
#include "calories_def.h"

// threads
#include "util_thread.h"

// ca
#include "ca.h"
#include "ca_custom.h"
#include "vectors.h"

// OpenGL
#if defined(__PLATFORM_MACOSX__)
  #include <GLUT/glut.h>
  #include <OpenGL/gl.h>
  #include <OpenGL/glu.h>
#else
  #include <GL/glut.h>
  #include <GL/gl.h>
  #include <GL/glu.h>
#endif

// oscpack
#include "osc/OscReceivedElements.h"
#include "osc/OscPacketListener.h"
#include "osc/OscOutboundPacketStream.h"
#include "ip/UdpSocket.h"

// STL
#include <iostream>
#include <string>
using namespace std;




//-----------------------------------------------------------------------------
// global variables and #defines
//-----------------------------------------------------------------------------


// osc port
long g_osc_port = 8888;
// ca dims
long g_ca_width = 8;
long g_ca_height = 8;
long g_ca_depth = 8;
long g_ca_dim = 3;
long g_x = 0, g_y = 0, g_z = 0;
long g_x2 = 0, g_y2 = 0, g_z2 = 0;
// the global ca
CA * g_ca = NULL;
long g_updates = 0;
long g_next = 0;
long g_active = 0;
long g_all = 0;
// width and height of the window
GLsizei g_width = 800;
GLsizei g_height = 600;
GLsizei g_last_width = g_width;
GLsizei g_last_height = g_height;
GLfloat g_size_x = 2.8;
GLfloat g_size_y = 2;
GLfloat g_size_z = 2;
// fill mode
GLenum g_fillmode = GL_FILL;
// light 0 position
GLfloat g_light0_pos[4] = { 2.0f, 1.2f, 4.0f, 1.0f };
// light 1 parameters
GLfloat g_light1_ambient[] = { .2f, .2f, .2f, 1.0f };
GLfloat g_light1_diffuse[] = { 1.0f, 1.0f, 1.0f, 1.0f };
GLfloat g_light1_specular[] = { 1.0f, 1.0f, 1.0f, 1.0f };
GLfloat g_light1_pos[4] = { -2.0f, 0.0f, -4.0f, 1.0f };
// modelview stuff
GLfloat g_angle_y = 0.0f;
GLfloat g_inc = 0.0f;
Vector3D g_eye_y( 0, 0, 1.0f );
Vector3D g_radius( 40.0f, 3.5f, 1.0f );
Vector3D g_bg;
GLfloat g_inc_val_mouse = .2f;
GLfloat g_linewidth = 2.0f;
// fullscreen
GLboolean g_fullscreen = FALSE;
// draw empty
GLboolean g_drawempty = FALSE;
// do
GLboolean g_fog = TRUE;
// auto
GLboolean g_auto = FALSE;
// render 3d cells
GLboolean g_render_cells = TRUE;
// text
char g_text[2048];
// position
Vector3D g_where;
// fog
GLuint fog_mode[4]; // Storage For Three/Four Types Of Fog
GLuint fog_filter;  // Which Fog To Use
GLfloat fog_density;  // Fog Density
// open sound control
UdpTransmitSocket * g_transmit = NULL;
osc::OutboundPacketStream * g_packet = NULL;
UdpTransmitSocket * g_transmit2 = NULL;
osc::OutboundPacketStream * g_packet2 = NULL;
const long SEND_SIZE = 4096;
char g_pbuffer[SEND_SIZE];
char g_pbuffer2[SEND_SIZE];
// hostname
string g_hostname = "localhost";
long g_port = 8889;
// mutex
XMutex g_mutex;




//-----------------------------------------------------------------------------
// function prototypes
//-----------------------------------------------------------------------------
void idleFunc( );
void displayFunc( );
void reshapeFunc( int width, int height );
void keyboardFunc( unsigned char, int, int );
void mouseFunc( int button, int state, int x, int y );
void initialize_graphics( );
void render_CA2D();
void sendCell( const Cell * cell );
void sendCell2( const Cell * cell );



//-----------------------------------------------------------------------------
// name: set3
// desc: ...
//-----------------------------------------------------------------------------
void set3( Vector3D & v, float v1, float v2, float v3 )
{
    v.x = v1;
    v.y = v2;
    v.z = v3;
}




//-----------------------------------------------------------------------------
// name: slew()
// desc: ...
//-----------------------------------------------------------------------------
void slew( Vector3D & v )
{
    v.x = (v.y - v.x) * v.z * .025 + v.x;
}



//-----------------------------------------------------------------------------
// name: CaloriesPacketListener
// desc: ...
//-----------------------------------------------------------------------------
class CaloriesPacketListener : public osc::OscPacketListener
{
protected:
    virtual void ProcessMessage( const osc::ReceivedMessage & m, 
                                 const IpEndpointName & remoteEndpoint )
    {
        try{
            if( !strcmp( m.AddressPattern(), "/calories/update" ) )
            {
                g_updates++;
            }
            else if( !strcmp( m.AddressPattern(), "/calories/get/next" ) )
            {
                g_next++;
            }
            else if( !strcmp( m.AddressPattern(), "/calories/get/next/active" ) )
            {
                g_active++;
            }
            else if( !strcmp( m.AddressPattern(), "/calories/get/next/all" ) )
            {
                g_all++;
            }
            else if( !strcmp( m.AddressPattern(), "/calories/get/cell/state" ) )
            {
                osc::ReceivedMessageArgumentStream oscin = m.ArgumentStream();
                osc::int32 x, y, z, value;
                oscin >> x >> y >> z;
                const Cell * cell = g_ca->getCell( x, y, z );
                // if NULL, ignore
                if( cell )
                {
                    sendCell2( cell );
                    // flag it
                    g_x2 = x; g_y2 = y; g_z2 = z;
                }
            }
            else if( !strcmp( m.AddressPattern(), "/calories/set/cell/state" ) )
            {
                osc::ReceivedMessageArgumentStream oscin = m.ArgumentStream();
                osc::int32 x, y, z, value;
                Cell * cell = NULL;
                oscin >> x >> y >> z >> value;
                // set it
                g_ca->setCell( x, y, z, value );
            }
        }catch( osc::Exception& e ){
            std::cerr << "error while parsing message: "
                << m.AddressPattern() << ": " << e.what() << "\n";
        }
    }
};




//-----------------------------------------------------------------------------
// name: cb_osc_listener()
// desc: ...
//-----------------------------------------------------------------------------
void * cb_osc_listener( void * )
{
    // instantiate listener
    CaloriesPacketListener listener;
    UdpListeningReceiveSocket s(
        IpEndpointName( IpEndpointName::ANY_ADDRESS, g_osc_port ),
        &listener );

    // print
    cerr << "OSC listener started on port: " << g_osc_port << "..." << endl;

    // go!
    s.RunUntilSigInt();

    return NULL;
}




//-----------------------------------------------------------------------------
// name: help()
// desc: ...
//-----------------------------------------------------------------------------
void help()
{
    fprintf( stderr, "----------------------------------------------------\n" );
    fprintf( stderr, "calories-3D (1.0.4)\n" );
    fprintf( stderr, "Ge Wang\n" );
    fprintf( stderr, "http://ccrma.stanford.edu/~ge/software/calories/\n" );
    fprintf( stderr, "----------------------------------------------------\n" );
    fprintf( stderr, "'h' - print this help message\n" );
    fprintf( stderr, "'0' - clear the cellular automaton\n" );
    fprintf( stderr, "[1-9] - randomize the automaton with various density\n" );
    fprintf( stderr, "'n' - compute next generation\n" );
    fprintf( stderr, "'N' - compute next generation + send all cells (use with care!)\n" );
    fprintf( stderr, "'M' - send all cells only (use with care!)\n" );
    fprintf( stderr, "'a' - toggle automatic evolution\n" );
    fprintf( stderr, "'[' and ']' - rotate automaton\n" );
    fprintf( stderr, "'-' and '+' - zoom away/closer to center of automaton\n" );
    fprintf( stderr, "'f' - toggle fog rendering (recommended)\n" );
    fprintf( stderr, "'<' and '>' - adjust fog density\n" );
    fprintf( stderr, "'s' - toggle fullscreen\n" );
    fprintf( stderr, "'r' - toggle wireframe rendering\n" );
    fprintf( stderr, "'q' - quit\n" );
    fprintf( stderr, "----------------------------------------------------\n" );
    fprintf( stderr, "\n" );
}




//-----------------------------------------------------------------------------
// name: usage()
// desc: ...
//-----------------------------------------------------------------------------
void usage()
{
    fprintf( stderr, "usage: calories  -[options]\n" );
    fprintf( stderr, "   [options] xyz:<N>|xy:<N>|x:<N>|y:<N>|z:<N>\n" );
    fprintf( stderr, "             hostname:<target>|port:<N>|help\n" );
    fprintf( stderr, "\n" );
    fprintf( stderr, "for example, to create a 32x32x32 cellular automaton, invoke:\n" );
    fprintf( stderr, "   calores -xyz:32\n" );
    fprintf( stderr, "\n" );
    fprintf( stderr, "calories version: 1.0.4\n" );
    fprintf( stderr, "\n" );

    exit(0);
}




//-----------------------------------------------------------------------------
// Name: main( )
// Desc: entry point
//-----------------------------------------------------------------------------
int main( int argc, char ** argv )
{
    // command line arguments
    for( int i = 1; i < argc; i++ )
    {
        if( !strncmp( argv[i], "-", 1 ) )
        {
            if( !strncmp(argv[i], "-x:", 3) )
                g_ca_width = atoi( argv[i]+3 ) > 0 ? atoi( argv[i]+3 ) : g_ca_width;
            else if( !strncmp(argv[i], "-y:", 3) )
                g_ca_height = atoi( argv[i]+3 ) > 0 ? atoi( argv[i]+3 ) : g_ca_height;
            else if( !strncmp(argv[i], "-z:", 3) )
                g_ca_depth = atoi( argv[i]+3 ) > 0 ? atoi( argv[i]+3 ) : g_ca_depth;
            else if( !strncmp(argv[i], "-xyz:", 5) )
                g_ca_width = g_ca_height = g_ca_depth = atoi( argv[i]+5 ) > 0 ? atoi( argv[i]+5 ) : g_ca_width;
            else if( !strncmp(argv[i], "-xy:", 4) )
            {
                g_ca_width = g_ca_height = atoi( argv[i]+4 ) > 0 ? atoi( argv[i]+4 ) : g_ca_width;
                g_ca_depth = 1; g_ca_dim = 2;
            }
            else if( !strncmp(argv[i], "-hostname:", 11) )
                g_hostname = argv[11];
            else if( !strncmp(argv[i], "-port:", 7) )
                g_port = atoi( argv[7] );
            else if( !strncmp(argv[i], "-help", 5) || !strncmp(argv[i], "--help", 6) )
                usage();
            else
            {
                fprintf( stderr, "invalid flag '%s'...\n", argv[i] );
                usage();
            }
        }
        else
        {
            usage();
        }
    }

#ifdef __OS_MACOSX__
    //save working dir
    char * cwd = getcwd( NULL, 0 );
#endif

    // compute size
    if( g_ca_width >= g_ca_height )
        g_size_y = g_size_x * (GLfloat)g_ca_height / g_ca_width;
    else
        g_size_x = g_size_y * (GLfloat)g_ca_width / g_ca_height;
    g_size_z = g_size_x * (GLfloat)g_ca_depth / g_ca_width;
    if( g_ca_depth > 1 ) g_ca_dim = 3;
    else { g_ca_dim = 2; g_ca_depth = 1; }

    // RALLY POINT: this is where we instantiate the CA we want to use
    g_ca = new CA_GOL();

    // start CA
    g_ca->init( g_ca_width, g_ca_height, g_ca_depth );

    // initialize GLUT
    glutInit( &argc, argv );

#ifdef __OS_MACOSX__
    //restore working dir
    chdir( cwd );
    free( cwd );
#endif

    // double buffer, use rgb color, enable depth buffer
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    // initialize the window size
    glutInitWindowSize( g_width, g_height );
    // set the window postion
    glutInitWindowPosition( 100, 100 );
    // create the window
    glutCreateWindow( "CAlories");
    // full screen
    if( g_fullscreen )
        glutFullScreen();

    // set the idle function - called when idle
    glutIdleFunc( idleFunc );
    // set the display function - called when redrawing
    glutDisplayFunc( displayFunc );
    // set the reshape function - called when client area changes
    glutReshapeFunc( reshapeFunc );
    // set the keyboard function - called on keyboard events
    glutKeyboardFunc( keyboardFunc );
    // set the mouse function - called on mouse stuff
    glutMouseFunc( mouseFunc );

    // do our own initialization
    initialize_graphics( );

    // print help
    help();

    // print
    fprintf( stderr, "initializing %dx%dx%d cellular automaton...\n", 
             g_ca_width, g_ca_height, g_ca_depth );
    // print network
    fprintf( stderr, "connecting to: '%s':%i...\n", g_hostname.c_str(), g_port );

    // initialize osc
    g_transmit = new UdpTransmitSocket( IpEndpointName( g_hostname.c_str(), g_port ) );
    g_packet = new osc::OutboundPacketStream( g_pbuffer, SEND_SIZE );
    g_transmit2 = new UdpTransmitSocket( IpEndpointName( g_hostname.c_str(), g_port ) );
    g_packet2 = new osc::OutboundPacketStream( g_pbuffer2, SEND_SIZE );

    // start listener
    XThread thread;
    thread.start( cb_osc_listener );

    // print
    fprintf( stderr, "\n" );

    // let GLUT handle the current thread from here
    glutMainLoop();

    return 0;
}




//-----------------------------------------------------------------------------
// Name: initialize_graphics( )
// Desc: sets initial OpenGL states and initializes any application data
//-----------------------------------------------------------------------------
void initialize_graphics()
{
    // set the GL clear color - use when the color buffer is cleared
    glClearColor( g_bg.x, g_bg.y, g_bg.z, 1.0f );
    glClearColor( 1, 1, 1, 1 );
    // set the shading model to 'smooth'
    glShadeModel( GL_SMOOTH );
    // enable depth
    glEnable( GL_DEPTH_TEST );
    // set the front faces of polygons
    glFrontFace( GL_CCW );
    // set fill mode
    glPolygonMode( GL_FRONT_AND_BACK, g_fillmode );
    // enable lighting
    glEnable( GL_LIGHTING );
    // enable lighting for front
    glLightModeli( GL_FRONT_AND_BACK, GL_TRUE );
    // material have diffuse and ambient lighting 
    glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
    // enable color
    glEnable( GL_COLOR_MATERIAL );
    // normalize (for scaling)
    glEnable( GL_NORMALIZE );
    // line width
    glLineWidth( g_linewidth );

    // enable light 0
    glEnable( GL_LIGHT0 );

    // setup and enable light 1
    glLightfv( GL_LIGHT1, GL_AMBIENT, g_light1_ambient );
    glLightfv( GL_LIGHT1, GL_DIFFUSE, g_light1_diffuse );
    glLightfv( GL_LIGHT1, GL_SPECULAR, g_light1_specular );
    glEnable( GL_LIGHT1 );

    // fog
    fog_mode[0] = 0; /*fog_mode[1] = GL_EXP; fog_mode[2] = GL_EXP2;*/ fog_mode[1] = GL_LINEAR;
    fog_filter = 0;
    fog_density = .08f;

    // fog
    GLfloat fogColor[4]= {1.0f, 1.0f, 1.0f, 1.0f};      // Fog Color
    if( fog_filter )
        glFogi(GL_FOG_MODE, fog_mode[fog_filter]);      // Fog Mode
    glFogfv(GL_FOG_COLOR, fogColor);            // Set Fog Color
    glFogf(GL_FOG_DENSITY, fog_density);                // How Dense Will The Fog Be
    glHint(GL_FOG_HINT, GL_DONT_CARE);          // Fog Hint Value
    glFogf(GL_FOG_START, 1.0f);             // Fog Start Depth
    glFogf(GL_FOG_END, 10.5f);               // Fog End Depth
    if( fog_filter )
        glEnable(GL_FOG);

    // check global flag
    if( g_fog )
    {
        glFogi(GL_FOG_MODE, fog_mode[fog_filter]);      // Fog Mode
        glEnable(GL_FOG);
    }
    else
        glDisable(GL_FOG);
}




//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( int w, int h )
{
    // save the new window size
    g_width = w; g_height = h;
    // map the view port to the client area
    glViewport( 0, 0, w, h );
    // set the matrix mode to project
    glMatrixMode( GL_PROJECTION );
    // load the identity matrix
    glLoadIdentity( );
    // create the viewing frustum
    gluPerspective( 45.0, (GLfloat) w / (GLfloat) h, .05, 50.0 );
    // set the matrix mode to modelview
    glMatrixMode( GL_MODELVIEW );
    // load the identity matrix
    glLoadIdentity( );
}




//-----------------------------------------------------------------------------
// Name: look( )
// Desc: ...
//-----------------------------------------------------------------------------
void look( )
{
    // position the view point
    gluLookAt( 0.0f, g_radius.x * sin( g_eye_y.x ), g_radius.x * cos( g_eye_y.x ), 
               0.0f, 0.0f, 0.0f, 
               0.0f, ( cos( g_eye_y.x ) < 0 ? -1.0f : 1.0f ), 0.0f );

    // set the position of the lights
    glLightfv( GL_LIGHT0, GL_POSITION, g_light0_pos );
    glLightfv( GL_LIGHT1, GL_POSITION, g_light1_pos );
}




//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
    switch( key )
    {
    case 'q':
        exit( 0 );
    break;
    case 's':
    {
        if( !g_fullscreen )
        {
            g_last_width = g_width;
            g_last_height = g_height;
            glutFullScreen();
        }
        else
            glutReshapeWindow( g_last_width, g_last_height );

        g_fullscreen = !g_fullscreen;
        fprintf( stderr, "[calories]: fullscreen:%s\n", g_fullscreen ? "ON" : "OFF" );
    }
    break;
    case 'f':
        g_fog = !g_fog;
        fprintf( stderr, "[calories]: fog:%s\n", g_fog ? "ON" : "OFF" );
        if( g_fog )
        {
            glFogi(GL_FOG_MODE, fog_mode[fog_filter]);      // Fog Mode
            glEnable(GL_FOG);
        }
        else
        {
            glDisable(GL_FOG);
        }
    break;
    case '<':
        fog_density *= .95f;
        fprintf( stderr, "[calories]: fog density:%f\n", fog_density );
        glFogf(GL_FOG_DENSITY, fog_density);
    break;
    case '>':
        fog_density *= 1.05f;
        fprintf( stderr, "[calories]: fog density:%f\n", fog_density );
        glFogf(GL_FOG_DENSITY, fog_density);
    break;
    case 'n':
        g_ca->update();
    break;
    case 'N':
        fprintf( stderr, "[calories]: updating and sending all states...\n" );
        g_ca->update();
        g_all++;
    break;
    case 'M':
        fprintf( stderr, "[calories]: sending all states...\n" );
        g_all++;
    break;
    case ']':
        g_eye_y.y -= .1f;
        //fprintf( stderr, "[calories]: yview:%f\n", g_eye_y.y );
    break;
    case '[':
        g_eye_y.y += .1f;
        //fprintf( stderr, "[calories]: yview:%f\n", g_eye_y.y );
    break;
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
        if( g_ca_dim < 3 ) g_ca->randomize( (key-'0') / 10.0 );
        else g_ca->randomize( (key-'0') / 100.0 );
    break;
    case 'e':
        g_drawempty = !g_drawempty;
    break;
    case '=':
        g_radius.y = .95 * g_radius.y;
        // fprintf( stderr, "[calories]: view radius:%f\n", g_radius.y );
    break;
    case '-':
        g_radius.y = 1.05 * g_radius.y;
        // fprintf( stderr, "[calories]: view radius:%f\n", g_radius.y );
    break;
    case 'r':
        g_render_cells = !g_render_cells;
        fprintf( stderr, "[calories]: render 3d cells:%s\n", g_render_cells ? "ON" : "OFF" );
    break;
    case 'a':
        g_auto = !g_auto;
        fprintf( stderr, "[calories]: auto update:%s\n", g_auto ? "ON" : "OFF" );
    break;
    case 'h':
        help();
    break;
    }

    // do a reshape since g_eye_y might have changed
    reshapeFunc( g_width, g_height );
    glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: mouseFunc( )
// Desc: handles mouse stuff
//-----------------------------------------------------------------------------
void mouseFunc( int button, int state, int x, int y )
{
    if( button == GLUT_LEFT_BUTTON )
    {
        // rotate
        if( state == GLUT_DOWN )
        {}
        else
        {}
    }
    else if ( button == GLUT_RIGHT_BUTTON )
    {
        if( state == GLUT_DOWN )
        {}
        else
        {}
    }
    else
        g_inc = 0.0f;

    glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
    // render the scene
    glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// name: sendCell()
// desc: ...
//-----------------------------------------------------------------------------
void sendCell( const Cell * cell )
{
    // start bundle
    g_packet->Clear();
    *g_packet << osc::BeginBundleImmediate << osc::BeginMessage( "/calories/cell/state" )
              << (int)cell->x << (int)cell->y << (int)cell->z << (int)cell->value << (float)cell->valuef << osc::EndMessage << osc::EndBundle;
    g_transmit->Send( g_packet->Data(), g_packet->Size() );
}




//-----------------------------------------------------------------------------
// name: sendCell2()
// desc: ...
//-----------------------------------------------------------------------------
void sendCell2( const Cell * cell )
{
    // start bundle
    g_packet2->Clear();
    *g_packet2 << osc::BeginBundleImmediate << osc::BeginMessage( "/calories/cell/state" )
              << (int)cell->x << (int)cell->y << (int)cell->z << (int)cell->value << (float)cell->valuef << osc::EndMessage << osc::EndBundle;
    g_transmit2->Send( g_packet2->Data(), g_packet2->Size() );
}




//-----------------------------------------------------------------------------
// name: draw_string()
// desc: ...
//-----------------------------------------------------------------------------
void draw_string( GLfloat x, GLfloat y, GLfloat z, const char * str, GLfloat scale = 1.0f )
{
    GLint len = strlen( str ), i;

    glPushMatrix();
    glTranslatef( x, y, z );
    glScalef( .001f * scale, .001f * scale, .001f * scale );
    for( i = 0; i < len; i++ )
        glutStrokeCharacter( GLUT_STROKE_ROMAN, str[i] );
    glPopMatrix();
}




//-----------------------------------------------------------------------------
// name: render_Cell2D()
// desc: ...
//-----------------------------------------------------------------------------
inline void render_Cell2D( const Cell * cell )
{
    GLfloat xs = g_size_x / g_ca_width / 2;
    int sel = ( cell->x == g_x && cell->y == g_y && cell->z == g_z );
    int sel2 = ( cell->x == g_x2 && cell->y == g_y2 && cell->z == g_z2 );

        // draw inner box
        if( sel ) glColor3f( .5f, .2f, 0.2f );
        else if( sel2 ) glColor3f( .25f, .25f, 1 );
        else glColor3f( .5f, .5f, .5f );
        glBegin( GL_LINE_LOOP );
            glTranslatef( 0, 0, .01 );
            glNormal3f( 0.0f, 0.0f, 1.0f );
            glVertex2f( xs, xs );
            glVertex2f( -xs, xs );
            glVertex2f( -xs, -xs );
            glVertex2f( xs, -xs );
        glEnd();

        if( cell->value && sel ) glColor3f( 1.0f, 1.0f, 0.0f );
        else if( cell->value && sel2 ) glColor3f( .5f, .5f, 1 );
        else glColor3f( 1.f*!cell->value, 1.f*!cell->value, 1.f*!cell->value );
        glBegin( GL_QUADS );
            glNormal3f( 0.0f, 0.0f, 1.0f );
            glVertex2f( xs, xs );
            glVertex2f( -xs, xs );
            glVertex2f( -xs, -xs );
            glVertex2f( xs, -xs );
        glEnd();

}




//-----------------------------------------------------------------------------
// name: render_CA2D()
// desc: ...
//-----------------------------------------------------------------------------
void render_CA2D()
{
    GLfloat xinc = g_size_x / g_ca_width;
    GLfloat yinc = g_size_y / g_ca_height;

    // loop over boxes
    for( long x = 0; x < g_ca_width; x++ )
    {
        for( long y = 0; y < g_ca_height; y++ )
        {
            glPushMatrix();
            glTranslatef( x * xinc, y * yinc, 0.0f );
            glPushMatrix();
            render_Cell2D( &(g_ca->grid[x][y][0]) );
            glPopMatrix();
            glPopMatrix();
        }
    }
}




//-----------------------------------------------------------------------------
// name: render_Cell3D()
// desc: ...
//-----------------------------------------------------------------------------
inline void render_Cell3D( const Cell * cell )
{
    GLfloat xs = g_size_x / g_ca_width;
    int sel = ( cell->x == g_x && cell->y == g_y && cell->z == g_z );
    int sel2 = ( cell->x == g_x2 && cell->y == g_y2 && cell->z == g_z2 );

    // go
    if( sel ) glColor3f( 1, 1, 0 );
    else if( sel2 ) glColor3f( .25f, .25f, 1 );
    else glColor3f( 0.9, 0.9, 0.9 );
    if( g_drawempty || cell->value || sel || sel2 ) glutWireCube( xs );
    if( sel ) glColor3f( .5, .5, 0 );
    else if( sel2 ) glColor3f( 0, 0, .75f );
    else glColor3f( 0.0, 0.5, 0 );
    if( (g_render_cells || sel || sel2) && (cell->value || cell->size > .005) )
        glutSolidCube( xs*cell->size );
}




//-----------------------------------------------------------------------------
// name: render_CA3D()
// desc: ...
//-----------------------------------------------------------------------------
void render_CA3D()
{
    GLfloat xinc = g_size_x / g_ca_width;
    GLfloat yinc = g_size_y / g_ca_height;
    GLfloat zinc = g_size_z / g_ca_depth;

    // loop over boxes
    for( long x = 0; x < g_ca_width; x++ )
    {
        for( long y = 0; y < g_ca_height; y++ )
        {
            for( long z = 0; z < g_ca_depth; z++ )
            {
                glPushMatrix();
                glTranslatef( x * xinc, y * yinc, z*zinc );
                glPushMatrix();
                render_Cell3D( &(g_ca->grid[x][y][z]) );
                glPopMatrix();
                glPopMatrix();
            }
        }
    }
}




//-----------------------------------------------------------------------------
// name: sendNext()
// desc: ...
//-----------------------------------------------------------------------------
bool sendNext( bool send0s = false )
{
    g_x++;
    if( g_x >= g_ca_width )
    {
        g_x = 0;
        g_y++;
        if( g_y >= g_ca_height )
        {
            g_y = 0;
            g_z++;
            if( g_z >= g_ca_depth )
                g_z = 0;
        }
    }

    // send
    Cell * cell = &g_ca->grid[g_x][g_y][g_z];
    if( send0s || cell->value )
    {
        sendCell( cell );
        return true;
    }

    return false;
}




//-----------------------------------------------------------------------------
// name: sendNextActive()
// desc: ...
//-----------------------------------------------------------------------------
void sendNextActive()
{
    if( g_ca->numActive == 0 ) return;
    long total = g_ca_width * g_ca_height * g_ca_depth;
    while( total-- && !sendNext() );
}




//-----------------------------------------------------------------------------
// name: sendAll()
// desc: ...
//-----------------------------------------------------------------------------
void sendAll()
{
    long total = g_ca_width * g_ca_height * g_ca_depth;
    while( total-- ) sendNext( true );
}




//-----------------------------------------------------------------------------
// Name: displayFunc( )
// Desc: callback function invoked to draw the client area
//-----------------------------------------------------------------------------
void displayFunc( )
{
    // clear the color and depth buffers
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    // save state
    glPushMatrix();

    // slew
    slew( g_eye_y ); slew( g_radius ); look();

    // size
    g_ca->updateLook();
    // update?
    if( g_updates || g_auto )
    {
        // go
        g_ca->update();
        if( g_updates > 0 ) g_updates--;
    }

    // send next
    if( g_next > 0 )
    {
        sendNext();
        g_next--;
    }

    // send next active
    if( g_active > 0 )
    {
        sendNextActive();
        g_active--;
    }

    // send all
    if( g_all > 0 )
    {
        sendAll();
        g_all--;
    }

    // draw CA
    glPushMatrix();
    if( g_ca_dim == 2 )
    {
        // center
        glTranslatef( -g_size_x/2+g_size_x/g_ca_width/2, 
                      -g_size_y/2+g_size_y/g_ca_height/2,
                      0 );
        // go
        render_CA2D();
    }
    else if( g_ca_dim == 3 )
    {
        // center
        glTranslatef( -g_size_x/2+g_size_x/g_ca_width/2, 
                      -g_size_y/2+g_size_y/g_ca_height/2,
                      -g_size_z/2+g_size_z/g_ca_depth/2 );
        // go
        render_CA3D();
    }
    glPopMatrix();

    // pop state
    glPopMatrix();

    // flush gl commands
    glFlush( );
    // swap the buffers
    glutSwapBuffers( );
}
